Event.observe(window, "load", function() {
  $('nojs').remove();
  document.body.appendChild(wholeApp());
});

/*** graphical components ***/

function wholeApp() {
  var t = randTable();
  var begin = new Ref(0);
  var consumers = new ListRef();
  var search = new Signal();
  var search_available = new Ref(true);
  var note = new OrderedMap();
  var note_reload = new Signal();

  return elem('div', {}, [
    positionControl(begin),
    columnSetControl(consumers, search_available, do_search),
    searchDisplay(search, search_available, begin.set.bind(begin)),
    tableDisplay(t, begin, consumers, note, note_reload),
    noteControl(note, note_reload)
  ]);

  function do_search(backward, fun, val) {
    var tt = t.clone();
    search.signal(
      begin.value() + (backward ? -1 : 1),
      backward,
      function(index){ return fun(tt, index).toString() == val; }
    );
  }
}

function positionControl(ref_begin) {
  var pbegin = new Ref(ref_begin.value());
  var amount = new Ref(10);
  return elem('p', {}, [
    unit([
      text('現在位置 = '),
      variableText(ref_begin)
    ]),
    unit([
      numberInputLong(pbegin, 0),
      text('へ'),
      disableableButton('移動', {}, new Ref().relate(pbegin, liftMaybe(function(val) {
        return function() { ref_begin.set(Math.floor(val)); };
      })))
    ]),
    unit([
      button('<<', {}, function() { ref_begin.set(ref_begin.value() - 1); }),
      button('>>', {}, function() { ref_begin.set(ref_begin.value() + 1); })
    ]),
    unit([
      disableableButton('<<<', {}, move(-1)),
      numberInput(amount, 10),
      disableableButton('>>>', {}, move(1))
    ])
  ]);

  function move(sign) {
    return new Ref().relate(amount, liftMaybe(function(val) {
      return function() { ref_begin.set(ref_begin.value() + sign * val); };
    }));
  }
}

function columnSetControl(ref_consumers, can_search, do_search) {
  var cols = new ListRef([
    {consumer: new Ref(function(t, i) { return i; })},
    {consumer: new Ref(function(t, i) { return t(i); })}
  ]);
  ref_consumers.relateEach(cols, prop('consumer'));
  var column_set = elem('div');
  add_column();
  return elem('div', {'class': 'columns'}, [
    column_set,
    elem('p', {}, [button('列を追加', {}, add_column)]).setStyle({textAlign: 'right'})
  ]);

  function add_column() {
    var col = { consumer: new Ref() };
    col.node = columnControl(col.consumer, can_search, do_search, remove_column.curry(col));
    column_set.appendChild(col.node);
    cols.push(col);
  }

  function remove_column(col) {
    column_set.removeChild(col.node);
    cols.remove(cols.value().indexOf(col));
  }
}

function columnControl(ref_consumer, can_search, do_search, kill_me) {
  var maybe_consumer = new Ref();
  var error = new Ref(null);
  maybe_consumer.onUpdate.add(function(val) {
    if(val[0]) {
      ref_consumer.set(val[1]);
      error.set(null);
    } else {
      error.set(val[1]);
    }
  });

  return elem('div', {'class': 'columnctl'}, [
    switcher(maybe_consumer, [
      ["銃/計算尺", gunDamage],
      ["ブレイカー", breakerDamage],
      ["その他武器/魔法", swordDamage],
      ["剰余", residue],
      ["ノックバック", knockback],
      ["連撃", combo],
      ["トレジャー", treasure]
    ]),
    searchControl(can_search, do_search, maybe_consumer),
    errorDisplay(error),
    elem('p', {}, [button('この列を削除', {}, kill_me)])
  ]);
}

function switcher(ref, choices) {
  var selected = new Ref();
  var node = elem('div', {}, [
    elem('p', {}, [select(selected, 0, choices.map(prop(0)))])
  ]);
  var child;
  var on_kill;
  switch_to(0);
  selected.onUpdate.add(function(index) {
    on_kill.signal();
    node.removeChild(child);
    switch_to(index);
  });
  return node;

  function switch_to(index) {
    on_kill = new Signal();
    child = choices[index][1](ref, on_kill);
    node.appendChild(child);
  }
}

function searchControl(can_search, do_search, maybe_consumer) {
  var value = new Ref();
  var trigger = new Ref().relate(can_search, maybe_consumer, function(ok, mcons) {
    if(!ok || !mcons[0])
      return null;
    return function(backward) {
      do_search(backward, mcons[1], value.value());
    };
  });
  return elem('p', {}, [
    disableableButton('<<前を検索', {}, new Ref().relate(trigger, liftMaybe(meth('curry', true)))),
    input(value, '0', Prototype.K, 10),
    disableableButton('次を検索>>', {}, new Ref().relate(trigger, liftMaybe(meth('curry', false))))
  ]);
}

function errorDisplay(error) {
  return variableNode('div', {}, new Ref().relate(error, function(errorval) {
    if(errorval === null)
      return elem('div');
    else
      return elem('p', {}, [text('error: ' + errorval)]);
  }));
}

function searchDisplay(sig_search, search_available, move_to) {
  var current = null;
  var status_text = new Ref('待機中');
  var extra_ctl = new Ref(elem('span'));
  sig_search.add(do_search);

  return elem('p', {}, [
    text('検索: '),
    variableText(status_text),
    variableNode('span', {}, extra_ctl)
  ]);

  function do_search(begin, backward, cond) {
    current = new Search(begin, backward, cond,
      function(index) { status_text.set('検索中(' + index + ')...'); },
      handle_complete,
      handle_suspend);
    start();
  }

  function handle_complete(index) {
    current = null;
    status_text.set('完了(' + index + ')');
    extra_ctl.set(elem('span'));
    move_to(index);
    search_available.set(true);
  }

  function handle_suspend(index) {
    status_text.set('停止(' + index + ')');
    extra_ctl.set(button('再開', {}, start));
    search_available.set(true);
  }

  function start() {
    search_available.set(false);
    extra_ctl.set(button('中断', {}, current.suspend.bind(current)));
    current.start();
  }
}

function tableDisplay(t, ref_begin, ref_consumers, note, note_reload) {
  var tbl = table(
    new Ref(50),
    new ListRef().asConcat(
      new ListRef([ruler]),
      new ListRef().asConcat(
        new ListRef().relateEach(ref_consumers, constant(text.curry('<>'))),
        new ListRef([notearea]))));
  var cols = new ListRef().relateEach(ref_consumers, function(consumer) {
    var col = {};
    col.remover = consumer.onUpdate.add(function(fun) {
      update_column(cols.value().indexOf(col), fun);
    });
    return col;
  });
  cols.onRemove.add(function(_, col) { col.remover(); });
  cols.onInsert.add(function(index) {
    update_column(index, ref_consumers.value()[index].value());
  });
  ref_begin.onUpdate.add(update_all);
  note_reload.add(update_notearea);
  update_all();
  return elem('div', {'class': 'tablebox'}, [tbl]);

  function update_all() {
    ref_consumers.value().each(function(ref, index) {
      update_column(index, ref.value());
    });
    update_notearea();
  }

  function update_column(index, fun) {
    var begin = ref_begin.value();
    $A(tbl.rows).each(function(row, i) {
      row.cells[index+1].firstChild.data = fun(t, begin + i).toString();
    });
  }

  function update_notearea() {
    var begin = ref_begin.value();
    var pos = ref_consumers.value().length + 1;
    $A(tbl.rows).each(function(row, i) {
      var d = note.lookup(begin + i);
      row.cells[pos].firstChild.value = d === undefined ? '' : d;
    });
  }

  function ruler(index) {
    return elem('span', {'class': 'ruler'}, [text(index.toString())]);
  }

  function notearea(index) {
    var text = new Ref();
    text.onUpdate.add(function(str) {
      var key = ref_begin.value() + index;
      if(str == '')
        note.remove(key);
      else
        note.insert(key, str);
    });
    return input(text, '', Prototype.K, 100);
  }
}

function noteControl(note, note_reload) {
  var ext = new Ref(elem('div'));
  return elem('div', {'class': 'notectl'}, [
    elem('p', {}, [
      unit([button('メモをテキストとして表示', {}, function() {
        ext.set(note_exporter());
      })]),
      unit([button('テキストからメモを導入', {}, function() {
        ext.set(note_importer());
      })])
    ]),
    variableNode('div', {}, ext)
  ]);

  function note_exporter() {
    return elem('pre', {'class': 'note'}, [text(serializeNote(note))]);
  }

  function note_importer() {
    var ta = elem('textarea', {rows: '10'});
    return elem('div', {}, [
      ta,
      elem('p', {}, [
        button('追加', {}, function() {
          ext.set(elem('div'));
          deserializeNote(note, ta.value);
          note_reload.signal();
        })
      ])
    ]);
  }
}

function serializeNote(note) {
  var s = '';
  note.each(function(o) {
    s += o.key + ':' + o.value + '\n';
  });
  return s;
}

function deserializeNote(note, data) {
  data = data.gsub(/\r\n/, '\n'); /* for IE */
  data.scan(/^(-?\d+):(.*)$/m, function(match) {
    var key = Number(match[1]);
    var current = note.lookup(key);
    if(current)
      note.insert(key, current + ' <> ' + match[2]);
    else
      note.insert(key, match[2]);
  });
}

/*** specific column types ***/

gunDamage = createRandConsumer([
    val("攻撃値", labelled("攻撃: ", numberInput), 38),
    val("倍率", labelled("倍率: ", numberInput), 3)
  ],
  function(a, r) {
    return function(t, i) {
      return limitDamage(Math.floor(rand98(t(i), a) * rand98(t(i+1), a) * r));
    };
  }
);

breakerDamage = createRandConsumer([
    val("攻撃値", labelled("攻撃: ", numberInput), 95),
    val("防御値", labelled("防御: ", numberInput), 0),
    val("倍率", labelled("倍率: ", numberInputLong), 17.3)
  ],
  function(a, d, r) {
    return function(t, i) {
      return limitDamage(Math.floor((rand98(t(i), a) - d) * rand11(t(i+1), r)));
    };
  }
);

swordDamage = createRandConsumer([
    val("攻撃/威力値", labelled("攻撃/威力: ", numberInput), 85),
    val("防御/魔防値", labelled("防御/魔防: ", numberInput), 0),
    val("倍率", labelled("倍率: ", numberInputLong), 36.9551)
  ],
  function(a, d, r) {
    return function(t, i) {
      return limitDamage(Math.floor((rand98(t(i), a) - d) * r));
    };
  }
);

residue = createRandConsumer([
    val("法", labelled("法: ", numberInputLong), 3)
  ],
  function(m) {
    return function(t, i) {
      return mod(t(i), m);
    };
  }
);

knockback = createRandConsumer([
    val("実行者レベル", labelled("実行者レベル: ", numberInput), 34),
    val("対象レベル", labelled("対象レベル: ", numberInput), 34),
    val("ノックバック値", labelled("ノックバック値: ", numberInput), 10)
  ],
  function(x, y, k) {
    return function(t, i) {
      var rate = Math.max(0, t(i) % x - t(i+1) % y + k);
      return t(i+2) % 100 < rate ? 'y' : 'n';
    };
  }
);

combo = createRandConsumer([
    val("連撃率", labelled("連撃率(%): ", numberInput), 3),
    val("HP割合", labelled("HP割合: ", select), 0, ["1/4 ～", "1/8 ～ 1/4", "1/16 ～ 1/8", "～ 1/16"])
  ],
  function(r, hp) {
    var k = [8, 16, 32, 64][hp];
    return function(t, i) {
      if(t(i) % 100 >= r) return 1;
      return Math.min(12, $R(1, 12).inject(2, function(n, z) {
        return n + (t(i+z) % 100 < k);
      }));
    };
  }
);

treasure = createRandConsumer([
    val("ギル率", labelled("ギル率(%): ", numberInput), 90),
    val("非レア率", labelled("非レア率(%): ", numberInput), 90),
    val("目標", select, 0, ["レアを狙う", "非レアを狙う"])
  ],
  function(gr, nrr, tgt) {
    return function(t, i) {
      var b = t(i) % 100 >= gr && (t(i+1) % 100 < nrr) == tgt;
      return b ? 'y' : 'n';
    };
  }
);

function rand11(rand, base) {
  if(base < 0.01) return 0;
  return (rand % Math.floor(100 * base)) / 100;
}

function rand98(rand, base) {
  return base + rand11(rand, base * 0.125);
}

function limitDamage(raw) {
  return Math.max(0, Math.min(9999, raw));
}

function createRandConsumer(spec, fun) {
  var initials = spec.findAll(prop('valued')).map(prop('defaultValue'));
  return function(ref, on_kill) {
    var inputs = [];
    var children = [];
    spec.each(function(sig) {
      if(sig.valued) {
        var cv = new Ref();
        children.push(sig.make(cv, initials[inputs.length]));
        inputs.push(cv);
      } else {
        children.push(sig.make());
      }
    });
    ref.relate.apply(ref, inputs.concat([liftError(fun)]));
    on_kill.add(save_as_defaults);
    return elem('p', {}, children);

    function save_as_defaults() {
      inputs.each(function(ref, index) {
        var v = ref.value();
        if(v[0])
          initials[index] = v[1];
      });
    }
  };
}

function cnode(proto) {
  return {
    valued : false,
    make : function() { return proto.cloneNode(true); }
  };
}

function val(desc, ctor, def) {
  var extra_args = $A(arguments).slice(3);
  return {
    valued : true,
    make : function(ref, val) {
      var iref = new Ref();
      ref.relate(iref, check);
      return ctor.apply(null, [iref, val].concat(extra_args));

      function check(t) {
        if(t === null)
          return [false, desc + 'が不正です'];
        else
          return [true, t];
      }
    },
    defaultValue : def
  };
}

/*** suspendable linear searcher ***/

function Search(begin, backward, cond, on_progress, on_complete, on_suspend) {
  this._pos = begin;
  this._backward = backward;
  this._cond = cond;
  this._on_progress = on_progress;
  this._on_complete = on_complete;
  this._on_suspend = on_suspend;
  this._to_suspend = false;
  this._running = false;
}

Search.prototype.start = function() {
  if(!this._running) {
    this._running = true;
    this._on_progress(this._pos);
    repeat(this._progress.bind(this));
  }
};

Search.prototype._progress = function() {
  if(this._to_suspend) {
    this._running = false;
    this._to_suspend = false;
    this._on_suspend(this._pos);
    return false;
  }

  var found = new Object;
  var interval = 321;
  var i;
  var self = this;
  var start_time = new Date().getTime();

  try {
    for(var m = 0; m < 7; m++) {
      if(this._backward)
        for(i = this._pos; i > this._pos - interval; i--)
          test(i);
      else
        for(i = this._pos; i < this._pos + interval; i++)
          test(i);
      this._pos = i;
      if(new Date().getTime() - start_time > 100)
        break;
    }
  } catch(e) {
      if(e == found) return false;
      throw e;
  }
  this._on_progress(i);
  return true;

  function test(i) {
    if(self._cond(i)) {
      self._running = false;
      self._pos = i;
      self._on_complete(i);
      throw found;
    }
  }
};

Search.prototype.suspend = function() {
  if(this._running)
    this._to_suspend = true;
};

function repeat(fun) {
  window.setTimeout(function(){
    if(fun()) repeat(fun);
  }, 0);
}

/*** associative container for ordered keys ***/

function OrderedMap() {
  this._val = [];
}

OrderedMap.prototype.lookup = function(key) {
  var r = this._findIndex(key);
  return r.exist ? this._val[r.pos].value : undefined;
};

OrderedMap.prototype._findIndex = function(key) {
  var first = 0, last = this._val.length;
  while(first != last) {
    var mid = div(first + last, 2);
    var mk = this._val[mid].key;
    if(mk == key)
      return {exist: true, pos: mid};
    else if(mk < key)
      first = mid + 1;
    else
      last = mid;
  }
  return {exist: false, pos: first};
};

OrderedMap.prototype.insert = function(key, value) {
  var r = this._findIndex(key);
  if(r.exist) {
    this._val[r.pos].value = value;
  } else {
    var x = this._val.slice(0, r.pos);
    x.push({key: key, value: value});
    this._val = x.concat(this._val.slice(r.pos));
  }
};

OrderedMap.prototype.remove = function(key) {
  var r = this._findIndex(key);
  if(r.exist)
    this._val = this._val.slice(0, r.pos).concat(this._val.slice(r.pos + 1));
};

OrderedMap.prototype.each = function(iter, ctx) {
  this._val.each(function(entry) {
    iter.call(ctx, Object.clone(entry));
  });
};

/*** Mersenne Twister implementation ***/

function randTable() {
  return randTableFrom(MT19937_1998.init(4537), 0);
}

function randTableFrom(initial, initial_page) {
  var current;
  var cache_forward;
  var cache_backward;
  var page;
  var page_start; /* optimization */
  reset();
  at.clone = function() { return randTableFrom(current, page); };
  return at;

  function reset() {
    current = initial;
    cache_forward = null;
    cache_backward = null;
    page = initial_page;
    page_start = page * MT19937_1998.N;
  }
  function at(index) {
    if(index < page_start || page_start + MT19937_1998.N <= index)
      move_to(div(index, MT19937_1998.N));
    return MT19937_1998.value(current, index - page_start);
  }
  function move_to(dest_page) {
    if(Math.abs(dest_page - page) > Math.abs(dest_page - initial_page) + 3)
      reset();
    while(page < dest_page)
      move_forward();
    while(page > dest_page)
      move_backward();
    page_start = page * MT19937_1998.N;
  }
  function move_forward() {
    cache_backward = current;
    current = cache_forward || MT19937_1998.next(current);
    cache_forward = null;
    page++;
  }
  function move_backward() {
    cache_forward = current;
    current = cache_backward || MT19937_1998.prev(current);
    cache_backward = null;
    page--;
  }
}

MT19937_1998 = {
  next : function(v) {
    var vec = v.clone();
    var nm = this.N - this.M;
    var matrix_a = this.MATRIX_A;
    for(var i = 0; i < nm; i++)
      upd(i, i+1, i+this.M);
    for(i = nm; i < this.N - 1; i++)
      upd(i, i+1, i-nm);
    upd(this.N-1, 0, this.M-1);
    return vec;
    function upd(i, j, k) {
      var y = vec[i] & 0x80000000 | vec[j] & 0x7fffffff;
      vec[i] = vec[k] ^ (y >>> 1) ^ (y & 1 ? matrix_a : 0);
    }
  },
  prev : function(v) {
    var vec = v.clone();
    var nm = this.N - this.M;
    var matrix_a = this.MATRIX_A;
    upd(this.N-1, -1, this.M-1);
    for(var i = this.N - 2; i >= nm; i--)
      upd(i, i+1, i-nm);
    for(i = nm - 1; i >= 0; i--)
      upd(i, i+1, i+this.M);
    upd(this.N-1, 0, this.M-1);
    return vec;
    function upd(i, j, k) {
      var z = vec[i] ^ vec[k];
      var y = z & 0x80000000
        ? (z ^ matrix_a) << 1 | 1
        : z << 1;
      if(j >= 0)
        vec[j] = vec[j] & 0x80000000 | y & 0x7fffffff;
      if(j != 0)
        vec[i] = y & 0x80000000 | vec[i] & 0x7fffffff;
    }
  },
  value : function(vec, index) {
    var x = vec[index];
    x ^= x >>> 11;
    x ^= (x << 7) & 0x9d2c5680;
    x ^= (x << 15) & 0xefc60000;
    x ^= x >>> 18;
    return x >>> 0;
  },
  init : function(seed) {
    var vec = new Array(this.N);
    for(var i = 0; i < this.N; i++) {
      seed ^= 0;
      vec[i] = seed;
      /* the multiplication below should not lose any precision,
       * because the result can be represented in at most 49 bits.
       */
      seed = 69069 * seed; 
    }
    return this.next(vec);
  },
  N : 624,
  M : 397,
  MATRIX_A : 0x9908b0df
};

/*** HTML building blocks ***/

function text(str) {
  return document.createTextNode(str);
}

function elem(tag, opt_attrs, children) {
  children = children || [];
  var r = new Element(tag, opt_attrs);
  for(var i = 0; i < children.length; i++) {
    // console.log('appending %dth child to %s: %s', i, tag, inspect(children[i]));
    r.appendChild(children[i]);
  }
  return r;
}

function variableText(ref) {
  var n = text(ref.value().toString());
  ref.onUpdate.add(function(val) {
    n.data = val.toString();
  });
  return n;
}

function button(value, attrs, onclick) {
  var n = elem('input', Object.extend({
    type: 'button',
    value: value,
    'class': 'button'
  }, attrs));
  n.onclick = onclick;
  return n;
}

function disableableButton(value, attrs, ref_trigger) {
  var n = button(value, attrs, function() {
    if(ref_trigger.value())
      ref_trigger.value()();
  });
  ref_trigger.onUpdate.add(function(trigger) {
    n.writeAttribute('disabled', trigger === null);
  });
  return n;
}

function input(ref, initial, make, size) {
  size = size === undefined ? '3' : size;
  var n = elem('input', {value: initial, size: size}, []);
  n.onkeyup = update;
  n.onmouseup = update;
  n.onchange = update;
  var prev;
  update();
  return n;
  function update() {
    if(n.value != prev) {
      prev = n.value;
      ref.set(make(n.value));
    }
  }
}

function numberInput(ref, initial, size) {
  return input(ref, initial.toString(), function(str) {
    var t = Number(str);
    return isNaN(t) || /^\s*$/.test(str) ? null : t;
  }, size);
}

function numberInputLong(ref, initial) {
  return numberInput(ref, initial, 10);
}

function select(ref, initial, choices) {
  var n = elem('select', {}, choices.map(option));
  n.selectedIndex = initial;
  update();
  n.onchange = update;
  return n;

  function update() {
    ref.set(n.selectedIndex);
  }
}

function option(str) {
  return elem('option', {}, [text(str)]);
}

function table(ref_length, ref_columns) {
  var n = elem('table');
  var cols = new ListRef();

  set_length(ref_length.value());
  ref_length.onUpdate.add(set_length);
  cols.onInsert.add(insert);
  cols.onRemove.add(remove);
  cols.relateEach(ref_columns, Prototype.K);
  return n;

  function set_length(len) {
    while(n.rows.length > len)
      n.deleteRow(-1);
    while(n.rows.length < len) {
      var index = n.rows.length;
      var row = n.insertRow(-1);
      cols.value().each(function(col) {
        row.insertCell(-1).appendChild(col(index));
      });
    }
  }

  function insert(pos, col) {
    $A(n.rows).each(function(row, index) {
      row.insertCell(pos).appendChild(col(index));
    });
  }

  function remove(pos) {
    $A(n.rows).each(function(row) {
      row.deleteCell(pos);
    });
  }
}

function unit(contents) {
  return elem('span', {}, [
    text(' '),
    elem('span', {'class': 'unit'}, contents),
    text(' ')
  ]);
}

function labelled(label, ctl) {
  return function() {
    return unit([
      text(label),
      ctl.apply(null, $A(arguments))
    ]);
  };
}

function variableNode(containerTag, containerAttrs, ref) {
  var child = ref.value();
  var n = elem(containerTag, containerAttrs, [child]);
  ref.onUpdate.add(function(new_child) {
    if(child != new_child) {
      n.replaceChild(new_child, child);
      child = new_child;
    }
  });
  return n;
}

/*** tools for reactive-like programming ***/

function Signal() {
  this._funs = [];
}

Signal.prototype.add = function(fn, ctx) {
  var entry = {fun: fn, context: ctx || null };
  this._funs.push(entry);
  return this._remove.bind(this, entry);
};

Signal.prototype._remove = function(e) {
  this._funs = this._funs.reject(function(entry) { return entry == e; });
}

Signal.prototype.connect = function(other) {
  return this.add(other.signal, other);
};

Signal.prototype.signal = function() {
  var args = $A(arguments);
  this._funs.each(function(entry) {
    entry.fun.apply(entry.context, args);
  });
};

function Ref(initial) {
  this._val = initial;
  this._initialized = initial !== undefined;
  this.onUpdate = new Signal();
}

Ref.prototype.set = function(val) {
  this._val = val;
  this._initialized = true;
  this.onUpdate.signal(val);
};

Ref.prototype.value = function() {
  return this._val;
};

Ref.prototype.initialized = function() {
  return this._initialized;
};

Ref.prototype.relate = function() {
  var self = this;
  var ref_ins = $A(arguments);
  var fun = ref_ins.pop();
  ref_ins.each(function(ref_in) { ref_in.onUpdate.add(update); });
  if(ref_ins.all(meth('initialized')))
    update();
  return self;
  
  function update() {
    self.set(
      fun.apply(null, ref_ins.map(meth('value'))));
  }
};

function ListRef(initial) {
  this._val = initial || [];
  this.onInsert = new Signal();
  this.onRemove = new Signal();
  this.onUpdate = new Signal();
}

ListRef.prototype.insert = function(pos, val) {
  this._val = this._val.slice(0, pos).concat([val]).concat(this._val.slice(pos));
  this.onInsert.signal(pos, val);
  this.onUpdate.signal(this._val);
};

ListRef.prototype.push = function(val) {
  this.insert(this._val.length, val);
};

ListRef.prototype.remove = function(pos) {
  var to_remove = this._val[pos];
  this._val = this._val.slice(0, pos).concat(this._val.slice(pos+1));
  this.onRemove.signal(pos, to_remove);
  this.onUpdate.signal(this._val);
};

ListRef.prototype.pop = function() {
  var r = this._val.last();
  this.remove(this._val.length - 1);
  return r;
}

ListRef.prototype.set = function(arr) {
  while(this._val.length > 0) {
    var x = this._val.pop();
    this.onRemove.signal(this._val.length, x);
  }
  arr.each((function(x, index) {
    this._val.push(x);
    this.onInsert.signal(index, x);
  }).bind(this));
  this.onUpdate.signal(this._val);
};

ListRef.prototype.value = function() { return this._val; };

ListRef.prototype.initialized = function() { return true; };

ListRef.prototype.relateEach = function(lref_in, fun) {
  var self = this;
  lref_in.onInsert.add(function(index, value) {
    self.insert(index, fun(value));
  });
  lref_in.onRemove.add(self.remove.bind(self));
  self.set(lref_in.value().map(fun));
  return self;
};

ListRef.prototype.relate = Ref.prototype.relate;

ListRef.prototype.asConcat = function(lref_x, lref_y) {
  var self = this;
  lref_x.onInsert.add(function(index, value) {
    self.insert(index, value);
  });
  lref_y.onInsert.add(function(index, value) {
    self.insert(lref_x.value().length + index, value);
  });
  lref_x.onRemove.add(function(index) {
    self.remove(index);
  });
  lref_y.onRemove.add(function(index) {
    self.remove(lref_x.value().length + index);
  });
  self.set(lref_x.value().concat(lref_y.value()));
  return self;
};

/*** utility functions ***/

function constant(x) {
  return function() { return x; };
}

function prop(name) {
  return function(obj) {
    return obj[name];
  };
}

function meth(name) {
  var args = $A(arguments).slice(1);
  return function(obj) {
    return obj[name].apply(obj, args);
  };
}

function div(x, y) {
  return Math.floor(x / y);
}

function mod(x, y) {
  return x - y * Math.floor(x / y);
}

function liftMaybe(fun) {
  return function() {
    var as = $A(arguments);
    if(as.any(function(a) { return a === null; }))
      return null;
    return fun.apply(null, as);
  };
}

function liftError(fun) {
  return function() {
    var as = $A(arguments);
    for(var i = 0; i < as.length; i++)
      if(!as[i][0])
        return as[i];
    return [true, fun.apply(null, as.map(prop(1)))];
  };
}

function inspect(obj) {
  if(obj && obj.nodeType == Node.TEXT_NODE)
    return 'text-node(' + obj.nodeValue + ')';
  else
    return Object.inspect(obj);
}
